"""
Unit tests for the tno.mpc.protocols.secure_comparison library.
"""

from __future__ import annotations

import asyncio

import pytest

from tno.mpc.encryption_schemes.dgk import DGK, DGKCiphertext
from tno.mpc.encryption_schemes.paillier import Paillier, PaillierCiphertext
from tno.mpc.encryption_schemes.utils import next_prime

from tno.mpc.protocols.secure_comparison import Initiator, KeyHolder, from_bits, to_bits
from tno.mpc.protocols.secure_comparison.test.conftest import COMPARISON_BIT_LENGTH

# Setting up the encryption schemes.
paillier = Paillier.from_security_parameter(key_length=1024)
paillier_n = paillier.public_key.n

# Sometimes we want full decryption to make checks.
u = next_prime(1 << (COMPARISON_BIT_LENGTH + 2))
dgk_full = DGK.from_security_parameter(v_bits=20, n_bits=128, u=u, full_decryption=True)

# Pairs (x, y) where x < y
_TEST_VALUES_SMALL_BIG = [
    (-400, -383),
    (-1, 0),
    (0, 2),
    (1, 10),
    (230, 269),
    (1508, 2408),
    (3122, 6048),
    (4250, 7804),
    (8668, 9015),
]
TEST_VALS_SMALLER_PLAINTEXT, TEST_VALS_BIGGER_PLAINTEXT = tuple(
    zip(*_TEST_VALUES_SMALL_BIG)
)
TEST_VALS_SMALLER = list(map(paillier.unsafe_encrypt, TEST_VALS_SMALLER_PLAINTEXT))
TEST_VALS_BIGGER = list(map(paillier.unsafe_encrypt, TEST_VALS_BIGGER_PLAINTEXT))
TEST_CONVERSION_TO_BOOL = [
    (TEST_VALS_SMALLER_PLAINTEXT[0], TEST_VALS_BIGGER_PLAINTEXT[0], True),
    (TEST_VALS_BIGGER_PLAINTEXT[0], TEST_VALS_SMALLER_PLAINTEXT[0], False),
    (TEST_VALS_BIGGER_PLAINTEXT[0], TEST_VALS_BIGGER_PLAINTEXT[0], True),
]
TEST_4A = [
    (0),
    (paillier_n),
    ((paillier_n - 1) // 2),
]
TEST_4D = [
    (to_bits(12345, COMPARISON_BIT_LENGTH), to_bits(23456, COMPARISON_BIT_LENGTH)),
    (to_bits(25, COMPARISON_BIT_LENGTH), to_bits(890, COMPARISON_BIT_LENGTH)),
]
TEST_5 = [
    (paillier.public_key.n, 0),
    (100, 0),
    (paillier.public_key.n, 1),
    (100, 1),
]
TEST_6 = [
    (0, paillier.unsafe_encrypt(0, apply_encoding=False)),
    (0, paillier.unsafe_encrypt(1, apply_encoding=False)),
    (1, paillier.unsafe_encrypt(0, apply_encoding=False)),
    (1, paillier.unsafe_encrypt(1, apply_encoding=False)),
]


def run_comparison(
    x_enc: PaillierCiphertext,
    y_enc: PaillierCiphertext,
    paillier_scheme: Paillier,
    dgk_scheme_not_full: DGK,
) -> PaillierCiphertext:
    """
    Perform integral comparison protocol for integration test.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input.
    :param paillier_scheme: Paillier scheme.
    :param dgk_scheme_not_full: DGK scheme.
    :return: Encryption of x <= y.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier_scheme)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier_scheme)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc_2 = KeyHolder.step_4a(
        z, dgk_scheme_not_full, paillier_scheme, COMPARISON_BIT_LENGTH
    )
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_scheme_not_full)
    d_enc = Initiator.step_4c(d_enc_2, r, dgk_scheme_not_full, paillier_scheme)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, alpha_tilde = Initiator.step_4e(
        r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier_scheme
    )
    w_is_enc = Initiator.step_4f(w_is_enc)
    s, delta_a = Initiator.step_4g()
    c_is_enc = Initiator.step_4h(
        s,
        alpha,
        alpha_tilde,
        d_enc,
        beta_is_enc,
        w_is_enc,
        delta_a,
        dgk_scheme_not_full,
    )
    c_is_enc = Initiator.step_4i(c_is_enc, dgk_scheme_not_full)
    delta_b = KeyHolder.step_4j(c_is_enc, dgk_scheme_not_full)
    zeta_1_enc, zeta_2_enc, delta_b_enc = KeyHolder.step_5(
        z, COMPARISON_BIT_LENGTH, delta_b, paillier_scheme
    )
    beta_lt_alpha_enc = Initiator.step_6(delta_a, delta_b_enc)
    x_leq_y_enc = Initiator.step_7(
        zeta_1_enc,
        zeta_2_enc,
        r,
        COMPARISON_BIT_LENGTH,
        beta_lt_alpha_enc,
        paillier_scheme,
    )
    return x_leq_y_enc


async def run_interactive_comparison(
    x: PaillierCiphertext | float,
    y: PaillierCiphertext | float,
    alice: Initiator,
    bob: KeyHolder,
) -> PaillierCiphertext:
    """
    Perform integral comparison protocol including communication
    between Initiator and KeyHolder for integration test.

    :param x: (Encryption of) first input.
    :param y: (Encryption of) second input.
    :param alice: Initiator of the secure comparison.
    :param bob: Keyholder in the secure comparison.
    :return: Encryption of x <= y.
    """
    task_initiator = asyncio.create_task(alice.perform_secure_comparison(x, y))
    task_keyholder = asyncio.create_task(bob.perform_secure_comparison())

    x_leq_y_enc, _ = await asyncio.gather(*(task_initiator, task_keyholder))
    return x_leq_y_enc


@pytest.mark.parametrize(
    "number",
    filter(
        lambda val: val >= 0, TEST_VALS_SMALLER_PLAINTEXT + TEST_VALS_BIGGER_PLAINTEXT  # type: ignore[operator, arg-type]
    ),
)
def test_bit_conversion(number: int) -> None:
    """
    Assert that conversion of a non-negative integer to bits and back
    returns the initial integer.

    :param number: Number to be converted.
    """
    assert from_bits(to_bits(number, COMPARISON_BIT_LENGTH)) == number


def test_modulo_n_squared() -> None:
    r"""
    Assert that the modulo reduction by $N^2$ is performed. It is not explicitly done
    in the code under test, but it is handled by the Paillier encryption.
    Note that the value of $y$ does not adhere to $0 \leq y < 2^l$. We chose the
    values of $x$ and $y$ such that z_enc needs modulo reduction.
    """
    l = COMPARISON_BIT_LENGTH
    x = 0
    x_enc = paillier.unsafe_encrypt(x, apply_encoding=False)
    y_enc = PaillierCiphertext(paillier.public_key.n_squared - 2, paillier)
    y = int(paillier.decrypt(y_enc, apply_encoding=False))
    r = paillier.public_key.n_squared

    z_enc = y_enc - x_enc + paillier.unsafe_encrypt((1 << l) + r, apply_encoding=False)
    z = (y - x + (1 << l) + r) % paillier.public_key.n_squared
    assert z == int(paillier.decrypt(z_enc, apply_encoding=False))


@pytest.mark.parametrize(
    "x, y, x_enc, y_enc",
    list(
        zip(
            TEST_VALS_SMALLER_PLAINTEXT,
            TEST_VALS_BIGGER_PLAINTEXT,
            TEST_VALS_SMALLER,
            TEST_VALS_BIGGER,
        )
    ),
)
def test_step_1(
    x: int, y: int, x_enc: PaillierCiphertext, y_enc: PaillierCiphertext
) -> None:
    r"""
    Assert that the computation in 'step_1' is equal to the alternative method
    of computation the authors give. In 'step_1', $[[z]] =
    [[y]] \cdot [[x]]^{-1} \cdot [[2^l + r]] = [[y]] - [[x]] + [[2^l + r]]$ is computed.
    Here, we compute $[[z]] \leftarrow [[y - x + 2^l + r]] \text{ mod } N^2$ and check
    whether it matches the approach above.

    :param x: First input as plaintext.
    :param y: Second input as plaintext.
    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    l = COMPARISON_BIT_LENGTH
    z_enc, r = Initiator.step_1(x_enc, y_enc, l, paillier)
    z_enc_test = paillier.unsafe_encrypt(
        (y - x + (1 << l) + r) % paillier.public_key.n_squared,
        apply_encoding=False,
    )
    assert paillier.decrypt(z_enc_test) == paillier.decrypt(z_enc)


def test_step_1_constraints_l_accepted_when_met() -> None:
    r"""
    Assert that valid constraints on $l$ are accepted by the function
    Initiator.step_1. The constraint, as stated in Protocol 3, is
    $l + 2 < log_2(N)$.
    """
    x = 1
    y = 2
    x_enc = paillier.unsafe_encrypt(x)
    y_enc = paillier.unsafe_encrypt(y)
    l = COMPARISON_BIT_LENGTH

    z_enc, r = Initiator.step_1(x_enc, y_enc, l, paillier)
    z_enc_test = paillier.unsafe_encrypt(
        (y - x + (1 << l) + r) % paillier.public_key.n_squared,
        apply_encoding=False,
    )
    assert paillier.decrypt(z_enc_test) == paillier.decrypt(z_enc)


def test_step_1_constraints_l_rejected_when_not_met() -> None:
    r"""
    Assert that invalid constraints on $l$ are rejected by the function
    Initiator.step_1. The constraint, as stated in Protocol 3, is
    $l + 2 < log_2(N)$.
    """
    l = paillier.public_key.n.bit_length() - 2
    x_enc = paillier.unsafe_encrypt(1)
    y_enc = paillier.unsafe_encrypt(2)

    with pytest.raises(AssertionError):
        Initiator.step_1(x_enc, y_enc, l, paillier)


@pytest.mark.parametrize(
    "x_enc, y_enc",
    list(
        zip(
            TEST_VALS_SMALLER,
            TEST_VALS_BIGGER,
        )
    ),
)
def test_step_2(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that the modulo reduction is performed correctly. We do this basically by checking
    whether $2^l$ is the same as $(1 << l)$.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    l = COMPARISON_BIT_LENGTH
    z_enc, _r = Initiator.step_1(x_enc, y_enc, l, paillier)
    z = int(paillier.decrypt(z_enc, apply_encoding=False))
    beta = z % 2**l
    z_test, beta_test = KeyHolder.step_2(z_enc, l, paillier)
    assert z_test == z and beta_test == beta


@pytest.mark.parametrize("z", TEST_4A)
def test_step_4a(z: int) -> None:
    r"""
    Assert that $d$ is computed based on the value of $z$
    and is encrypted.

    :param z: Plaintext result of step 1, here hand-picked.
    """

    d = bool(z < (paillier_n - 1) // 2)

    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    assert d == dgk_full.decrypt(d_enc)


@pytest.mark.parametrize(
    "x_enc, y_enc",
    list(
        zip(
            TEST_VALS_SMALLER,
            TEST_VALS_BIGGER,
        )
    ),
)
def test_step_4b(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that $\beta$ is transformed correctly into bits and the bitwise
    encryption is correct.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, _ = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    _, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    beta_bin_raw = bin(beta)[2:]
    beta_bin = [int(bit) for bit in beta_bin_raw]
    beta_bin.reverse()
    while len(beta_bin) != COMPARISON_BIT_LENGTH:
        beta_bin.append(0)

    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    for index, bit in enumerate(beta_is_enc):
        assert dgk_full.decrypt(bit) == int(beta_bin[index])


@pytest.mark.parametrize(
    "d_enc, r",
    [
        (dgk_full.unsafe_encrypt(0, apply_encoding=False), 0),
        (dgk_full.unsafe_encrypt(1, apply_encoding=False), 0),
        (dgk_full.unsafe_encrypt(0, apply_encoding=False), 100),
        (dgk_full.unsafe_encrypt(1, apply_encoding=False), 100),
    ],
)
def test_step_4c_sets_to_zero(
    d_enc: DGKCiphertext,
    r: int,
) -> None:
    r"""
    The value of $[d]$ is set to zero if $r$ is smaller than half the paillier
    scheme modulus.

    :param d_enc: Encrypted bit.
    :param r: The randomness value $r$ from step 1. Here the value is
        hand-picked.
    """
    d_enc_corrected = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    assert dgk_full.decrypt(d_enc_corrected) == 0


@pytest.mark.parametrize(
    "d_enc, r, test_result",
    [
        (dgk_full.unsafe_encrypt(0, apply_encoding=False), paillier_n - 100, 0),
        (dgk_full.unsafe_encrypt(1, apply_encoding=False), paillier_n - 100, 1),
    ],
)
def test_step_4c_maintains_original_value(
    d_enc: DGKCiphertext,
    r: int,
    test_result: int,
) -> None:
    r"""
    The value of $[d]$ maintains its value if $r$ is at least half the
        paillier scheme modulus.

    :param d_enc: Encrypted bit.
    :param r: The randomness value $r$ from step 1. Here the value is
        hand-picked.
    :param test_result: The expected output of 'step_4c', $[d]$.
    """
    d_enc_corrected = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    assert dgk_full.decrypt(d_enc_corrected) == test_result


@pytest.mark.parametrize("r", [-10, paillier_n + 10])
def test_step_4c_raises_valueerror_for_inappropriate_randomness(r: int) -> None:
    r"""
    The value of $[d]$ is corrected based on the value of $r$. 'step_4c'
    should give an assertion error if the constraint $0 \leq r < (N-1) \div 2$
    is not met.

    :param r: The provided, incorrect, randomness.
    """
    d_enc = dgk_full.unsafe_encrypt(0, apply_encoding=False)

    with pytest.raises(AssertionError):
        Initiator.step_4c(d_enc, r, dgk_full, paillier)


@pytest.mark.parametrize("alpha, beta", TEST_4D)
def test_step_4d(alpha: list[int], beta: list[int]) -> None:
    r"""
    Assert that $[\alpha_i \oplus \beta_i]$ is computed correctly.
    Here we compute the value using the unencrypted values and check
    whether the decryption of the encrypted output of 'step_4d' is the same.

    :param alpha: Computed value in 'step_3' by A, here hand-picked.
    :param beta: Computed value in 'step_2' by B, here hand-picked.
    """
    beta_is_enc = [
        dgk_full.unsafe_encrypt(beta[i], apply_encoding=False)
        for i in range(COMPARISON_BIT_LENGTH)
    ]
    for i in range(COMPARISON_BIT_LENGTH):
        alpha_xor_beta = alpha[i] ^ beta[i]
        alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
        assert alpha_xor_beta == dgk_full.decrypt(alpha_is_xor_beta_is_enc[i])


@pytest.mark.parametrize("x_enc, y_enc", list(zip(TEST_VALS_SMALLER, TEST_VALS_BIGGER)))
def test_step_4e(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that the value of $[w_i]$ is set correctly. We compute this slightly
    different than in 'step_4e', namely we make use of the multiplicative inverse and
    homomorphism to reduce the equation in a different manner.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    d_enc = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, _ = Initiator.step_4e(r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier)
    alpha_tilde_test = (r - paillier_n) % (1 << COMPARISON_BIT_LENGTH)
    alpha_tilde_test_bits = to_bits(alpha_tilde_test, COMPARISON_BIT_LENGTH)
    beta_bits = to_bits(beta, COMPARISON_BIT_LENGTH)
    d = int(dgk_full.decrypt(d_enc))

    for i in range(COMPARISON_BIT_LENGTH):
        if alpha[i] == alpha_tilde_test_bits[i]:
            assert dgk_full.decrypt(w_is_enc[i]) == alpha[i] ^ beta_bits[i]
        else:
            assert dgk_full.decrypt(w_is_enc[i]) == alpha[i] ^ beta_bits[i] - d


@pytest.mark.parametrize("x_enc, y_enc", list(zip(TEST_VALS_SMALLER, TEST_VALS_BIGGER)))
def test_step_4f(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    We basically test whether multiplication by a bit $i$ that is left-shifted
    every iteration is the same as multiplcation by $2^i$. We can't check whether
    $[w_i]^{2^i}$ works, as exponentation of PaillierCiphertexts is not implemented.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    d_enc = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, _ = Initiator.step_4e(r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier)
    w_test = [
        int(dgk_full.decrypt(w_is_enc[i])) * (2**i) % dgk_full.public_key.u
        for i in range(COMPARISON_BIT_LENGTH)
    ]
    w_is_enc = Initiator.step_4f(w_is_enc)
    for i in range(COMPARISON_BIT_LENGTH):
        assert w_test[i] == int(dgk_full.decrypt(w_is_enc[i], apply_encoding=False))


@pytest.mark.parametrize("x_enc, y_enc", list(zip(TEST_VALS_SMALLER, TEST_VALS_BIGGER)))
def test_step_4h(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that the value of $[c_i]$ is computed correctly. We do this by calculating
    it slightly different, namely $[c_i] = s + \alpha_i
    d*(\tilde{\alpha}_i-\alpha_i) - \beta_i] + 3*(\Sigma^{l-1}_{j=i+1}[w_j]) \mod n$
    instead of $[c_i] = [s] \cdot [\alpha_i] \cdot
    [d]^{\tilde{\alpha}_i-\alpha_i} \cdot [\beta_i]^{-1} \cdot (\Pi^{l-1}_{j=i+1}[w_j])^3 \mod n$.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    d_enc = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, alpha_tilde = Initiator.step_4e(
        r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier
    )
    w_is_enc = Initiator.step_4f(w_is_enc)
    s, delta_a = Initiator.step_4g()
    c_is_enc = Initiator.step_4h(
        s, alpha, alpha_tilde, d_enc, beta_is_enc, w_is_enc, delta_a, dgk_full
    )
    d = int(dgk_full.decrypt(d_enc))
    beta_bits = to_bits(beta, COMPARISON_BIT_LENGTH)

    for i in range(COMPARISON_BIT_LENGTH):
        temp = int(
            s
            + alpha[i]
            + (alpha_tilde[i] - alpha[i]) * d
            - beta_bits[i]
            + 3 * sum(int(dgk_full.decrypt(wj)) for wj in w_is_enc[i + 1 :])
        )
        assert (
            temp % dgk_full.public_key.u
            == int(dgk_full.decrypt(c_is_enc[i + 1])) % dgk_full.public_key.u
        )


@pytest.mark.parametrize("x_enc, y_enc", list(zip(TEST_VALS_SMALLER, TEST_VALS_BIGGER)))
def test_step_4j(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that $\delta_B$ is set correctly. If any of the values in $c_i$
    decrypts to 0, $\delta_B$ should be 1. We check this by iterating over $[c_i]$
    for all $0 \leq i < l$ and checking the result with the result of 'step_4i'.
    However, the resulting list of that step is one element too long: there is an
    element prepended, which we leave out of consideration here.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    d_enc = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, alpha_tilde = Initiator.step_4e(
        r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier
    )
    w_is_enc = Initiator.step_4f(w_is_enc)
    s, delta_a = Initiator.step_4g()
    c_is_enc = Initiator.step_4h(
        s, alpha, alpha_tilde, d_enc, beta_is_enc, w_is_enc, delta_a, dgk_full
    )
    c_is_enc = Initiator.step_4i(c_is_enc, dgk_full)
    delta_b = KeyHolder.step_4j(c_is_enc, dgk_full)

    for i in range(COMPARISON_BIT_LENGTH):
        if dgk_full.is_zero(
            c_is_enc[i + 1]
        ):  # Plus 1 here to correct the element added in Initiator.step_4h
            assert delta_b == 1


@pytest.mark.parametrize("z, delta_b", TEST_5)
def test_step_5(z: int, delta_b: int) -> None:
    r"""
    We compute $z \div 2^l$ slightly different than in 'step_5'. Namely
    we have eliminated the if-statement and have used a masking technique.

    :param z: Result of 'step_1', here hand-picked.
    :param delta_b: Bit result of 'step_4j'. Here the value is hand-picked.
    """
    l = COMPARISON_BIT_LENGTH
    n = paillier.public_key.n
    zeta_1_enc, zeta_2_enc, delta_b_enc = KeyHolder.step_5(
        z, COMPARISON_BIT_LENGTH, delta_b, paillier
    )
    assert paillier.decrypt(zeta_1_enc) == z // (2**l)
    assert paillier.decrypt(zeta_2_enc) == int(z < (n - 1) // 2) * (
        (z + n) // (2**l)
    ) + int(z >= (n - 1) // 2) * (z // (2**l))
    assert paillier.decrypt(delta_b_enc) == delta_b


@pytest.mark.parametrize("delta_a, delta_b_enc", TEST_6)
def test_step_6(delta_a: int, delta_b_enc: PaillierCiphertext) -> None:
    r"""
    Test whether $[[\beta < \alpha]]$ is computed correctly. We
    have eliminated the if-statement and have used a masking technique.
    We also used the value of $\delta_B$ directly, instead of using the
    encrypted value as should be used in the protocol.

    :param delta_a: Bit value computed in 'step_4g', here this is hand-picked.
    :param delta_b_enc: Encrypted bit computed in 'step_5', here this is hand-picked.
    """
    beta_lt_alpha_enc = Initiator.step_6(delta_a, delta_b_enc)
    delta_a_true_part = delta_a * delta_b_enc
    decryption_delta_b_enc = int(paillier.decrypt(delta_b_enc))
    delta_a_false_part = int(not delta_a) * (1 - decryption_delta_b_enc)
    assert paillier.decrypt(beta_lt_alpha_enc) == paillier.decrypt(
        delta_a_true_part + delta_a_false_part
    )


@pytest.mark.parametrize("x_enc, y_enc", list(zip(TEST_VALS_SMALLER, TEST_VALS_BIGGER)))
def test_step_7(x_enc: PaillierCiphertext, y_enc: PaillierCiphertext) -> None:
    r"""
    Assert that the computation of $x \leq y$ is correct. In the function 'step_7'
    the multiplicative inverse is used to negate an encrypted integer: $[-x] = [x]^{-1}$.
    This comes in handy, as we cannot compute the multiplication of two PaillierCiphertexts.
    However, here we can just work with the unencrypted values, such that we can multiply
    them.
    (Note: the fix mentioned in 'step_7' about $\zeta_1$ and $\zeta_2$ is not used here)

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input, satisfying x_enc < y_enc.
    """
    z_enc, r = Initiator.step_1(x_enc, y_enc, COMPARISON_BIT_LENGTH, paillier)
    z, beta = KeyHolder.step_2(z_enc, COMPARISON_BIT_LENGTH, paillier)
    alpha = Initiator.step_3(r, COMPARISON_BIT_LENGTH)
    d_enc = KeyHolder.step_4a(z, dgk_full, paillier, COMPARISON_BIT_LENGTH)
    beta_is_enc = KeyHolder.step_4b(beta, COMPARISON_BIT_LENGTH, dgk_full)
    d_enc = Initiator.step_4c(d_enc, r, dgk_full, paillier)
    alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)
    w_is_enc, alpha_tilde = Initiator.step_4e(
        r, alpha, alpha_is_xor_beta_is_enc, d_enc, paillier
    )
    w_is_enc = Initiator.step_4f(w_is_enc)
    s, delta_a = Initiator.step_4g()
    c_is_enc = Initiator.step_4h(
        s, alpha, alpha_tilde, d_enc, beta_is_enc, w_is_enc, delta_a, dgk_full
    )
    c_is_enc = Initiator.step_4i(c_is_enc, dgk_full)
    delta_b = KeyHolder.step_4j(c_is_enc, dgk_full)
    zeta_1_enc, zeta_2_enc, delta_b_enc = KeyHolder.step_5(
        z, COMPARISON_BIT_LENGTH, delta_b, paillier
    )
    beta_lt_alpha_enc = Initiator.step_6(delta_a, delta_b_enc)
    x_leq_y_enc = Initiator.step_7(
        zeta_1_enc,
        zeta_2_enc,
        r,
        COMPARISON_BIT_LENGTH,
        beta_lt_alpha_enc,
        paillier,
    )
    x_leq_y = (z // (2**COMPARISON_BIT_LENGTH)) - (
        (r // (2**COMPARISON_BIT_LENGTH)) + (int(beta < from_bits(alpha)))
    ) % paillier.public_key.n_squared

    assert x_leq_y == paillier.decrypt(x_leq_y_enc)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "x, y", list(zip(TEST_VALS_SMALLER_PLAINTEXT, TEST_VALS_BIGGER_PLAINTEXT))
)
async def test_smaller_than(
    x: int,
    y: int,
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Assert that the secure comparison returns the expected result when the
    input satisfies a "smaller than" relation, interactive and non-interactive.

    :param x: First input.
    :param y: Second input, satisfying x < y.
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication
    x_enc = paillier_strict.unsafe_encrypt(x)
    y_enc = paillier_strict.unsafe_encrypt(y)
    dgk_full_ = keyholder.scheme_dgk

    x_leq_y_enc = run_comparison(x_enc, y_enc, paillier_strict, dgk_full_)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 1

    x_leq_y_enc = await run_interactive_comparison(x_enc, y_enc, initiator, keyholder)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "x, y", list(zip(TEST_VALS_BIGGER_PLAINTEXT, TEST_VALS_SMALLER_PLAINTEXT))
)
async def test_greater_than(
    x: int,
    y: int,
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Assert that the secure comparison returns the expected result when the
    input satisfies a "greater than" relation, interactive and non-interactive.

    :param x: First input.
    :param y: Second input, satisfying x > y.
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication
    x_enc = paillier_strict.unsafe_encrypt(x)
    y_enc = paillier_strict.unsafe_encrypt(y)
    dgk_full_ = keyholder.scheme_dgk

    x_leq_y_enc = run_comparison(x_enc, y_enc, paillier_strict, dgk_full_)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 0

    x_leq_y_enc = await run_interactive_comparison(x_enc, y_enc, initiator, keyholder)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 0


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "x, y", list(zip(TEST_VALS_BIGGER_PLAINTEXT, TEST_VALS_BIGGER_PLAINTEXT))
)
async def test_equal_to(
    x: int,
    y: int,
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Assert that the secure comparison returns the expected result when the
    input satisfies a "is equal to" relation, interactive and non-interactive.

    :param x: First input.
    :param y: Second input, satisfying x == y.
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication
    x_enc = paillier_strict.unsafe_encrypt(x)
    y_enc = paillier_strict.unsafe_encrypt(y)
    dgk_full_ = keyholder.scheme_dgk

    x_leq_y_enc = run_comparison(x_enc, y_enc, paillier_strict, dgk_full_)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 1

    x_leq_y_enc = await run_interactive_comparison(x_enc, y_enc, initiator, keyholder)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("x, y, boolean_result", TEST_CONVERSION_TO_BOOL)
async def test_conversion_to_boolean(
    x: int,
    y: int,
    boolean_result: bool,
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Test conversion from decrypted result to boolean. If this test fails,
    the issue is probably in a depencency; however, the users of this package
    expect that the decrypted result works properly as a boolean and so we
    should know. Tests for interactive and non-interactive.

    :param x: First input.
    :param y: Second input.
    :param boolean_result: Boolean result of the comparison of x_enc and
        y_enc, e.g. (x <= y).
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication
    x_enc = paillier_strict.unsafe_encrypt(x)
    y_enc = paillier_strict.unsafe_encrypt(y)
    dgk_full_ = keyholder.scheme_dgk

    x_leq_y_enc = run_comparison(x_enc, y_enc, paillier_strict, dgk_full_)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert bool(x_leq_y) == boolean_result

    x_leq_y_enc = await run_interactive_comparison(x_enc, y_enc, initiator, keyholder)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert bool(x_leq_y) == boolean_result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "x, y", list(zip(TEST_VALS_SMALLER_PLAINTEXT, TEST_VALS_BIGGER_PLAINTEXT))
)
async def test_smaller_than_unencrypted_inputs(
    x: int,
    y: int,
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Assert that the secure comparison returns the expected result when the
    input satisfies a "smaller than" relation, interactive.

    :param x: First input.
    :param y: Second input, satisfying x < y.
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication

    x_leq_y_enc = await run_interactive_comparison(x, y, initiator, keyholder)
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)
    assert x_leq_y == 1


@pytest.mark.asyncio
async def test_parallel_runs(
    paillier_strict: Paillier,
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    Test multiple runs of `perform_secure_comparison simultaneously.

    :param paillier_strict: Paillier scheme.
    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    initiator, keyholder = playerpair_with_dict_communication

    x_enc_1 = paillier_strict.unsafe_encrypt(11)
    x_enc_2 = paillier_strict.unsafe_encrypt(41)
    y_enc_1 = paillier_strict.unsafe_encrypt(12)
    y_enc_2 = paillier_strict.unsafe_encrypt(41)

    task_alice_1 = asyncio.create_task(
        initiator.perform_secure_comparison(x_enc_1, y_enc_1)
    )
    task_alice_2 = asyncio.create_task(
        initiator.perform_secure_comparison(x_enc_2, y_enc_2)
    )
    task_bob_1 = asyncio.create_task(keyholder.perform_secure_comparison())
    task_bob_2 = asyncio.create_task(keyholder.perform_secure_comparison())

    x_leq_y_1, _, x_leq_y_2, _ = await asyncio.gather(
        *(task_alice_1, task_bob_1, task_alice_2, task_bob_2)
    )

    assert paillier_strict.decrypt(x_leq_y_1) == 1 == paillier_strict.decrypt(x_leq_y_2)


@pytest.mark.asyncio
async def test_if_different_schemes_then_raises_valueerror(
    playerpair_with_dict_communication: tuple[Initiator, KeyHolder],
) -> None:
    """
    A value error should be raised when the paillier or DGK scheme that is received conflicts with the readily initialized scheme.

    :param playerpair_with_dict_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    # pylint: disable=protected-access
    initiator, keyholder = playerpair_with_dict_communication
    initiator._scheme_paillier = paillier

    task_initiator = asyncio.create_task(initiator.receive_encryption_schemes())
    task_keyholder = asyncio.create_task(keyholder.make_and_send_encryption_schemes())

    with pytest.raises(ValueError, match=".*Paillier"):
        await asyncio.gather(*(task_initiator, task_keyholder))

    initiator._scheme_paillier = keyholder._scheme_paillier
    initiator._scheme_dgk = dgk_full

    task_initiator = asyncio.create_task(initiator.receive_encryption_schemes())
    task_keyholder = asyncio.create_task(keyholder.make_and_send_encryption_schemes())

    with pytest.raises(ValueError, match=".*DGK"):
        await asyncio.gather(*(task_initiator, task_keyholder))



"""
Implementation of secure comparison protocol as given in https://eprint.iacr.org/2018/1100.pdf.
"""

# Explicit re-export of all functionalities, such that they can be imported properly. Following
# https://www.python.org/dev/peps/pep-0484/#stub-files and
# https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport

from tno.mpc.protocols.secure_comparison.communicator import (
    Communicator as Communicator,
)
from tno.mpc.protocols.secure_comparison.initiator import Initiator as Initiator
from tno.mpc.protocols.secure_comparison.keyholder import KeyHolder as KeyHolder
from tno.mpc.protocols.secure_comparison.utils import from_bits as from_bits
from tno.mpc.protocols.secure_comparison.utils import to_bits as to_bits

__version__ = "4.4.0"



"""Initiator of protocol, i.e. performs step 1. Alice; A in the paper."""

from __future__ import annotations

from secrets import choice, randbelow
from typing import Any, cast

from tno.mpc.encryption_schemes.dgk import DGK, DGKCiphertext
from tno.mpc.encryption_schemes.paillier import Paillier, PaillierCiphertext

from tno.mpc.protocols.secure_comparison.communicator import Communicator
from tno.mpc.protocols.secure_comparison.utils import to_bits


class Initiator:
    """
    Player Alice in the secure comparison protocol, initiates.
    """

    def __init__(
        self,
        l_maximum_bit_length: int,
        communicator: Communicator | None = None,
        other_party: str = "",
        scheme_paillier: Paillier | None = None,
        scheme_dgk: DGK | None = None,
        session_id: int = 0,
    ) -> None:
        r"""
        :param l_maximum_bit_length: maximum bit length used to constrain variables ($l$).
        :param communicator: object for handling communication with the KeyHolder during the protocol.
        :param other_party: identifier of the other party
        :param scheme_paillier: Paillier encryption scheme (without secret key).
        :param scheme_dgk: DGK encryption scheme (without secret key).
        :param session_id: keeps track of the session.
        """

        self.l_maximum_bit_length = l_maximum_bit_length
        self.communicator = communicator
        self.other_party = other_party
        self._scheme_paillier = scheme_paillier
        self._scheme_dgk = scheme_dgk
        self.session_id = session_id

    @property
    def scheme_paillier(self) -> Paillier:
        """
        Paillier scheme of the initiator.

        :raise ValueError: No scheme available.
        :return: Paillier scheme.
        """
        if self._scheme_paillier is None:
            raise ValueError("No Paillier scheme has been initialized or received.")
        return self._scheme_paillier

    @property
    def scheme_dgk(self) -> DGK:
        """
        DGK scheme of the initiator.

        :raise ValueError: No scheme available.
        :return: DGK scheme.
        """
        if self._scheme_dgk is None:
            raise ValueError("No DGK scheme has been initialized or received.")
        return self._scheme_dgk

    async def perform_secure_comparison(
        self,
        x: PaillierCiphertext | float,
        y: PaillierCiphertext | float,
    ) -> PaillierCiphertext:
        r"""
        Performs all steps of the secure comparison protocol for Alice.
        Performs required communication with Bob.

        :param x: first (encrypted) input variable $x$ or $[[x]]$.
        :param y: second (encrypted) input variable $y$ or $[[y]]$.
        :return: Encrypted value of $(x<=y)$: $[[(x<=y)]]$.
        :raise ValueError: raised when communicator is not properly configured.
        """
        if self.communicator is None:
            raise ValueError("Communicator not properly initialized.")

        self.session_id += 1
        session_id = self.session_id

        # make sure you have the schemes. Always receive them to make sure they are the same.
        await self.receive_encryption_schemes(session_id)
        self._start_randomness_generation()

        x_enc = (
            x
            if isinstance(x, PaillierCiphertext)
            else self.scheme_paillier.unsafe_encrypt(x)
        )
        y_enc = (
            y
            if isinstance(y, PaillierCiphertext)
            else self.scheme_paillier.unsafe_encrypt(y)
        )

        # step 1
        z_enc, r_plain = Initiator.step_1(
            x_enc, y_enc, self.l_maximum_bit_length, self.scheme_paillier
        )

        z_enc.randomize()
        await self.communicator.send(
            self.other_party, z_enc, msg_id=f"step_1_session_{session_id}"
        )

        # step 3
        alpha = Initiator.step_3(r_plain, self.l_maximum_bit_length)

        d_enc, beta_is_enc = await self.communicator.recv(
            self.other_party, msg_id=f"step_4b_session_{session_id}"
        )

        # step 4c
        d_enc = Initiator.step_4c(d_enc, r_plain, self.scheme_dgk, self.scheme_paillier)

        # step 4d
        alpha_is_xor_beta_is_enc = Initiator.step_4d(alpha, beta_is_enc)

        # step 4e
        w_is_enc_step4e, alpha_tilde = Initiator.step_4e(
            r_plain, alpha, alpha_is_xor_beta_is_enc, d_enc, self.scheme_paillier
        )

        # step 4f
        w_is_enc = Initiator.step_4f(w_is_enc_step4e)

        # step 4g
        s_plain, delta_a = Initiator.step_4g()

        # step 4h
        c_is_enc_step4h = Initiator.step_4h(
            s_plain,
            alpha,
            alpha_tilde,
            d_enc,
            beta_is_enc,
            w_is_enc,
            delta_a,
            self.scheme_dgk,
        )

        # step 4i
        c_is_enc = Initiator.step_4i(c_is_enc_step4h, self.scheme_dgk, do_shuffle=True)

        for c in c_is_enc:
            c.randomize()
        await self.communicator.send(
            self.other_party, c_is_enc, msg_id=f"step_4i_session_{session_id}"
        )
        zeta_1_enc, zeta_2_enc, delta_b_enc = await self.communicator.recv(
            self.other_party, msg_id=f"step_5_session_{session_id}"
        )

        # step 6
        beta_lt_alpha_enc = Initiator.step_6(delta_a, delta_b_enc)

        # step 7
        x_leq_y_enc = Initiator.step_7(
            zeta_1_enc,
            zeta_2_enc,
            r_plain,
            self.l_maximum_bit_length,
            beta_lt_alpha_enc,
            self.scheme_paillier,
        )

        return x_leq_y_enc

    async def receive_encryption_schemes(self, session_id: int = 1) -> None:
        """
        Receives encryption schemes Paillier and DGK (without secret keys) from Bob.

        :param session_id: distinguish communication different sessions.
        :raise ValueError: raised when communicator is not properly configured.
        :raise ValueError: raised when received scheme is different from readily
            initialized scheme.
        """
        if self.communicator is None:
            raise ValueError("Communicator not properly initialized.")

        received_scheme_paillier, received_scheme_dgk = await self.communicator.recv(
            self.other_party, msg_id=f"schemes_session_{session_id}"
        )
        if self._scheme_paillier is None:
            self._scheme_paillier = received_scheme_paillier
        elif self.scheme_paillier != received_scheme_paillier:
            raise ValueError(
                "Readily available Paillier scheme and received Paillier scheme are different."
            )
        if self._scheme_dgk is None:
            self._scheme_dgk = received_scheme_dgk
        elif self.scheme_dgk != received_scheme_dgk:
            raise ValueError(
                "Readily available DGK scheme and received DGK scheme are different."
            )

    def _start_randomness_generation(self) -> None:
        """
        Boot randomness generation in both encryption schemes.
        """
        self.scheme_paillier.boot_randomness_generation(1)
        self.scheme_dgk.boot_randomness_generation(self.l_maximum_bit_length + 1)

    @staticmethod
    def shuffle(values: list[Any]) -> list[Any]:
        r"""
        Shuffle the list in random order.

        :param values: List of objects that is to be shuffled.
        :return: Shuffled version of the input list.
        """
        values = values.copy()
        shuffled_values = []
        while len(values):
            this_value = choice(values)
            values.remove(this_value)
            shuffled_values.append(this_value)
        return shuffled_values

    @staticmethod
    def step_1(
        x_enc: PaillierCiphertext,
        y_enc: PaillierCiphertext,
        l: int,
        scheme_paillier: Paillier,
    ) -> tuple[PaillierCiphertext, int]:
        r"""
        $A$ chooses a random number $r, 0 \leq r < N$, and computes
        $$[[z]] \leftarrow [[y - x + 2^l + r]] = [[x]] \cdot [[y]]^{-1} \cdot [[2^l + r]]
        \mod N^2.$$

        :param x_enc: Encrypted value of $x$: $[[x]]$.
        :param y_enc: Encrypted value of $y$: $[[y]]$.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x$, $y$ that will be given as
            input to this method.
        :param scheme_paillier: Paillier encryption scheme.
        :return: Tuple containing as first entry the encrypted value of $z$:
            $[[z]] \leftarrow [[y - x + 2^l + r]] = [[y]] \cdot [[x]]^{-1} \cdot [[2^l + r]] \mod
            N^2$. The second entry is the randomness value $r$.
        """
        assert (1 << (l + 2)) < scheme_paillier.public_key.n // 2
        r = randbelow(scheme_paillier.public_key.n)
        # Note: the paper has a typo here, it says x - y, i/o y - x.

        return (
            y_enc
            - x_enc
            + scheme_paillier.unsafe_encrypt((1 << l) + r, apply_encoding=False),
            r,
        )

    @staticmethod
    def step_3(r: int, l: int) -> list[int]:
        r"""
        $A$ computes $\alpha = r \mod 2^l$.

        :param r: The randomness value $r$ from step 1.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x, y$ that will
            be given as input to this method.
        :return: Value $\alpha = r \mod 2^l$ as bits.
        """
        return to_bits(r % (1 << l), l)

    @staticmethod
    def step_4c(
        d_enc: DGKCiphertext, r: int, scheme_dgk: DGK, scheme_paillier: Paillier
    ) -> DGKCiphertext:
        r"""
        $A$ corrects $[d]$ by setting $[d] \leftarrow [0]$ whenever $0 \leq r < (N - 1)/2$.

        :param d_enc: Encrypted value of $d$: $[d]$.
        :param r: The randomness value $r$ from step 1.
        :param scheme_dgk: DGK encryption scheme.
        :param scheme_paillier: Paillier encryption scheme.
        :return: Corrected encrypted value of $d$: $[d]$. If $0 \leq r < (N - 1)/2$, then
            $[d] \leftarrow [0]$, else $[d]$ remains unaltered.
        """
        assert (
            0 <= r < scheme_paillier.public_key.n
        )  # If step 1 is used, this is no issue. But this function can also be called separately.
        if r < (scheme_paillier.public_key.n - 1) // 2:
            d_enc = scheme_dgk.unsafe_encrypt(0, apply_encoding=False)
        return d_enc

    @staticmethod
    def step_4d(
        alpha: list[int], beta_is_enc: list[DGKCiphertext]
    ) -> list[DGKCiphertext]:
        r"""
        For each $i, 0 \leq i < l$, $A$ computes $[\alpha_i \oplus \beta_i]$ as follows:
        if $\alpha_i = 0$ then $[\alpha_i \oplus \beta_i] \leftarrow [\beta_i]$ else
        $[\alpha_i \oplus \beta_i] \leftarrow [1] \cdot [\beta_i]^{-1} \mod n$.

        :param alpha: The value $\alpha$ from step 3.
        :param beta_is_enc: List containing the encrypted values of $\beta_i$:
            $[\beta_i], 0 \leq i < l$.
        :return: List containing the encrypted values of the bits
            $\alpha_i \oplus \beta_i$: $[\alpha_i \oplus \beta_i], 0 \leq i < l$.
        """

        def compute_xor(alpha_i: int, beta_i_enc: DGKCiphertext) -> DGKCiphertext:
            r"""
            Compute $[\alpha_i \oplus \beta_i]$.

            :param alpha_i: The $i$-th bit of $\alpha$: $\alpha_i$.
            :param beta_i_enc: The encrypted $i$-th bit of $\beta$: $[\beta_i]$.
            :return: Encrypted value of $\alpha_i \oplus \beta_i$: $[\alpha_i \oplus \beta_i]$.
            """
            if alpha_i == 0:
                return beta_i_enc
            # else (alpha_i == 1)
            return 1 - beta_i_enc

        return list(
            map(
                compute_xor,
                alpha,
                beta_is_enc,
            )
        )

    @staticmethod
    def step_4e(
        r: int,
        alpha: list[int],
        alpha_is_xor_beta_is_enc: list[DGKCiphertext],
        d_enc: DGKCiphertext,
        scheme_paillier: Paillier,
    ) -> tuple[list[DGKCiphertext], list[int]]:
        r"""
        A computes $\tilde{\alpha} = (r - N) \mod 2^l$, the corrected value of $\alpha$ in case a
        carry-over actually did occur and adjusts $[\alpha_i \oplus \beta_i]$ for each $i$:
        If $\alpha_i = \tilde{\alpha}_i$ then $[w_i] \leftarrow [\alpha_i \oplus \beta_i]$
        else $[w_i] \leftarrow [\alpha_i \oplus \beta_i] \cdot [d]^{-1} \mod n$

        :param r: The randomness value $r$ from step 1.
        :param alpha: The value $\alpha$ from step 3.
        :param alpha_is_xor_beta_is_enc: List containing the encrypted values of the bits
            $\alpha_i \oplus \beta_i$: $[\alpha_i \oplus \beta_i], 0 \leq i < l$.
        :param d_enc: Encrypted value of $d$: $[d]$.
        :param scheme_paillier: Paillier encryption scheme.
        :return: Tuple containing as first entry a list containing the encrypted values of the bits
            $w_i$: $[w_i], 0 \leq i < l$.
            The second entry is the value $\tilde{\alpha} = (r - N) \mod 2^l$ as bits.
        """
        l = len(alpha_is_xor_beta_is_enc)

        def compute_w(
            alpha_i: int, alpha_tilde_i: int, alpha_i_xor_beta_i_enc: DGKCiphertext
        ) -> DGKCiphertext:
            r"""
            Compute $[w_i]$.

            :param alpha_i: The $i$-th bit of $\alpha$: $\alpha_i$.
            :param alpha_tilde_i: The $i$-th bit of $\tilde{\alpha}$: $\tilde{\alpha}_i$.
            :param alpha_i_xor_beta_i_enc: Encrypted value of the bit $\alpha_i \oplus \beta_i$:
                $[\alpha_i \oplus \beta_i]$.
            :return: Encrypted value of $w_i$: $[w_i]$.
            """
            if alpha_i == alpha_tilde_i:
                return alpha_i_xor_beta_i_enc
            # else
            return alpha_i_xor_beta_i_enc - d_enc

        alpha_tilde = to_bits(int((r - scheme_paillier.public_key.n) % (1 << l)), l)
        return (
            list(
                map(
                    compute_w,
                    alpha,
                    alpha_tilde,
                    alpha_is_xor_beta_is_enc,
                )
            ),
            alpha_tilde,
        )

    @staticmethod
    def step_4f(w_is_enc: list[DGKCiphertext]) -> list[DGKCiphertext]:
        r"""
        For each $i, 0 \leq i < l$, $A$ computes $[w_i] \leftarrow [w_i]^{2^i} \mod n$
        such that these values will not interfere each other when added.

        :param w_is_enc: List containing the encrypted values of the bits $w_i$: $[w_i],
            0 \leq i < l$.
        :return: List containing the encrypted values of the bits $w_i$: $[w_i], 0 \leq i < l$.
        """
        base = 1

        def compute_w(w_i_enc: DGKCiphertext) -> DGKCiphertext:
            r"""
            Compute $[w_i]$.

            :param w_i_enc: Encrypted value of $w_i$: $[w_i]$.
            :return: Encrypted value of $w_i$: $[w_i]$.
            """
            nonlocal base
            w_i_enc = w_i_enc * base
            base <<= 1
            return w_i_enc

        return list(map(compute_w, w_is_enc))

    @staticmethod
    def step_4g() -> tuple[int, int]:
        r"""
        $A$ chooses a uniformly random bit $\delta_A$ and computes $s = 1 - 2 \cdot \delta_A$.

        :return: Tuple containing as first entry the value $s = 1 - 2 \cdot \delta_A$.
            The second entry is the value $\delta_A$.
        """
        delta_a = randbelow(2)
        return 1 - 2 * delta_a, delta_a

    @staticmethod
    def step_4h(
        s: int,
        alpha: list[int],
        alpha_tilde: list[int],
        d_enc: DGKCiphertext,
        beta_is_enc: list[DGKCiphertext],
        w_is_enc: list[DGKCiphertext],
        delta_a: int,
        scheme_dgk: DGK,
    ) -> list[DGKCiphertext]:
        r"""
        For each $i, 0 \leq i < l$, $A$ computes $[c_i] = [s] \cdot [\alpha_i] \cdot
        [d]^{\tilde{\alpha}_i-\alpha_i} \cdot [\beta_i]^{-1} \cdot
        (\Pi^{l-1}_{j=i+1}[w_j])^3 \mod n$.
        We add an additional value $[c_{-1}]$, with
        $c_{-1}=\delta_A + \Sigma^{l-1}_{i=0}(x_i \oplus y_i)$ to also make
        the scheme work in case of equality of $x$ and $y$.

        :param s: The value $s$ from step 4g.
        :param alpha: The value $\alpha$ from step 3.
        :param alpha_tilde: The value $\tilde{\alpha}$ from step 4e.
        :param d_enc: Encrypted value of $d$: $[d]$.
        :param beta_is_enc: List containing the encrypted values of the bits $\beta_i$:
            $[\beta_i], 0 \leq i < l$.
        :param w_is_enc: List containing the encrypted values of the bits $w_i$: $[w_i],
            0 \leq i < l$.
        :param delta_a: The value $\delta_A$ from step 4g.
        :param scheme_dgk: DGK encryption scheme.
        :return: List containing the encrypted values of the bits $c_i$:
            $[c_i] = [s] \cdot [\alpha_i]
            \cdot [d]^{\tilde{\alpha}_i-\alpha_i} \cdot [\beta_i]^{-1}
            \cdot (\Pi^{l-1}_{j=i+1}[w_j])^3 \mod n, 0 \leq i < l$.
        """
        l = len(beta_is_enc)

        c_is_enc = [
            scheme_dgk.unsafe_encrypt(s, apply_encoding=False) for _ in range(l)
        ]
        w_is_enc_sum: int | DGKCiphertext = 0

        # pre-compute 3 options for d_enc * (alpha_i, alpha_tilde_i) to improve efficiency
        d_enc_mult_table = {
            -1: d_enc * -1,
            0: d_enc * 0,
            1: d_enc * 1,
        }

        for i, alpha_i, alpha_tilde_i in zip(
            range(l - 1, -1, -1),
            reversed(alpha),
            reversed(alpha_tilde),
        ):
            c_is_enc[i] += (
                int(alpha_i)
                + d_enc_mult_table[alpha_tilde_i - alpha_i]
                - beta_is_enc[i]
                + 3 * w_is_enc_sum
            )
            w_is_enc_sum += w_is_enc[i]
        # we use here the fix from the paper to make equality work
        c_is_enc.insert(0, cast(DGKCiphertext, delta_a + w_is_enc_sum))
        return c_is_enc

    @staticmethod
    def step_4i(
        c_is_enc: list[DGKCiphertext], scheme_dgk: DGK, do_shuffle: bool = True
    ) -> list[DGKCiphertext]:
        r"""
        $A$ blinds the numbers $c_i$ by raising them to a random non-zero exponent
        $r_i \in \{1,\ldots,u-1\}$.

        :param c_is_enc: List containing the encrypted values of the bits $c_i$: $[c_i],
            0 \leq i < l$.
        :param scheme_dgk: DGK encryption scheme.
        :param do_shuffle: Boolean parameter stating whether or not the bits should
            be shuffled randomly.
        :return: List containing the encrypted values of the masked bits $c_i$: $[c_i],
            0 \leq i < l$.
        """
        u = scheme_dgk.public_key.u

        def mask(c_i_enc: DGKCiphertext) -> DGKCiphertext:
            r"""
            Compute $[c_i]$.

            :param c_i_enc: Encrypted value of the bit $c_i$: $[c_i]$.
            :return: Encrypted value of the bit $c_i$: $[c_i]$.
            """
            c_i_enc *= randbelow(u - 1) + 1
            return c_i_enc

        c_is_enc_masked = list(map(mask, c_is_enc))
        return Initiator.shuffle(c_is_enc_masked) if do_shuffle else c_is_enc_masked

    @staticmethod
    def step_6(delta_a: int, delta_b_enc: PaillierCiphertext) -> PaillierCiphertext:
        r"""
        $A$ computes $[[(\beta < \alpha)]]$ as follows: if $\delta_A = 1$ then
        $[[(\beta < \alpha)]] \leftarrow [[\delta_B]]$ else
        $[[(\beta < \alpha)]] \leftarrow [[1]] \cdot [[\delta_B]]^{-1} \mod N^2$.

        :param delta_a: The value $\delta_A$ from step 4g.
        :param delta_b_enc: Encrypted value of $\delta_B$: $[[\delta_B]]$.
        :return: Encrypted value of $(\beta < \alpha)$: $[[(\beta < \alpha)]]$.
        """
        if delta_a == 1:
            return delta_b_enc
        return 1 - delta_b_enc

    @staticmethod
    def step_7(
        zeta_1_enc: PaillierCiphertext,
        zeta_2_enc: PaillierCiphertext,
        r: int,
        l: int,
        beta_lt_alpha_enc: PaillierCiphertext,
        scheme_paillier: Paillier,
    ) -> PaillierCiphertext:
        r"""
        $A$ computes $[[(x \leq y)]] \leftarrow
        [[\zeta]] \cdot ([[ r \div 2^l]] \cdot [[(\beta < \alpha)]])^{-1} \mod N^2$, where
        $\zeta = \zeta_1$, if $r < (N - 1) / 2$, else $\zeta = \zeta_2$.

        :param zeta_1_enc: Encrypted value of $\zeta_1$: $[[\zeta_1]]$.
        :param zeta_2_enc: Encrypted value of $\zeta_2$: $[[\zeta_2]]$.
        :param r: The randomness value $r$ from step 1.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x, y$ that will be given
            as input to this method.
        :param beta_lt_alpha_enc: Encrypted value of $(\beta < \alpha)$: $[[(\beta < \alpha)]]$.
        :param scheme_paillier: Paillier encryption scheme.
        :return: Encrypted value of $(x \leq y)$: $[[(x \leq y)]]$. This is the final result of the
            computation.
        """
        # We use the fix mentioned in the paper here, to also make this work in case of overflow
        zeta_enc = (
            zeta_1_enc if r < (scheme_paillier.public_key.n - 1) // 2 else zeta_2_enc
        )
        return zeta_enc - (
            scheme_paillier.unsafe_encrypt(r // (1 << l), apply_encoding=False)
            + beta_lt_alpha_enc
        )



"""
Protocol for a communicator object.
"""

from typing import Any, Protocol


class Communicator(Protocol):
    async def send(self, party_id: str, message: Any, msg_id: str) -> None: ...

    async def recv(self, party_id: str, msg_id: str) -> Any: ...



"""Party that holds the secret keys. Bob; B in the paper."""

from __future__ import annotations

from typing import cast

from tno.mpc.encryption_schemes.dgk import DGK, DGKCiphertext
from tno.mpc.encryption_schemes.paillier import Paillier, PaillierCiphertext
from tno.mpc.encryption_schemes.utils import next_prime

from tno.mpc.protocols.secure_comparison.communicator import Communicator
from tno.mpc.protocols.secure_comparison.utils import to_bits


class KeyHolder:
    """
    Player Bob in the secure comparison protocol, holds the keys.
    """

    def __init__(
        self,
        l_maximum_bit_length: int,
        communicator: Communicator | None = None,
        other_party: str = "",
        scheme_paillier: Paillier | None = None,
        scheme_dgk: DGK | None = None,
        session_id: int = 0,
    ) -> None:
        r"""
        :param l_maximum_bit_length: maximum bit length used to constrain variables ($l$).
        :param communicator: object for handling communication with the Initiator during the protocol.
        :param other_party: identifier of the other party
        :param scheme_paillier: Paillier encryption scheme (including secret key)
            used to produce $[[x]]$ and $[[y]]$, Alice's input.
        :param scheme_dgk: DGK encryption scheme (including secret key).
        :param session_id: keeps track of the session.
        """

        self.l_maximum_bit_length = l_maximum_bit_length
        self.communicator = communicator
        self.other_party = other_party
        self._scheme_paillier = scheme_paillier
        self._scheme_dgk = scheme_dgk
        self.session_id = session_id

    @property
    def scheme_paillier(self) -> Paillier:
        """
        Paillier scheme of the keyholder.

        :raise ValueError: No scheme available.
        :return: Paillier scheme.
        """
        if self._scheme_paillier is None:
            raise ValueError("No Paillier scheme has been initialized or received.")
        return self._scheme_paillier

    @property
    def scheme_dgk(self) -> DGK:
        """
        DGK scheme of the keyholder.

        :raise ValueError: No scheme available.
        :return: DGK scheme.
        """
        if self._scheme_dgk is None:
            raise ValueError("No DGK scheme has been initialized or received.")
        return self._scheme_dgk

    async def perform_secure_comparison(self) -> None:
        """
        Performs the secure comparison secure comparison for Bob.
        Including required communication with Alice.

        :raise ValueError: raised when communicator is not properly configured.
        """
        if self.communicator is None:
            raise ValueError("Communicator not properly initialized.")

        self.session_id += 1
        session_id = self.session_id

        # make sure you have the schemes and share the public keys
        await self.make_and_send_encryption_schemes(session_id)
        self._start_randomness_generation()

        z_enc = await self.communicator.recv(
            self.other_party, msg_id=f"step_1_session_{session_id}"
        )

        # step 2
        z_plain, beta = KeyHolder.step_2(
            z_enc, self.l_maximum_bit_length, self.scheme_paillier
        )

        # step 4a
        d_enc = KeyHolder.step_4a(
            z_plain, self.scheme_dgk, self.scheme_paillier, self.l_maximum_bit_length
        )

        # step 4b
        beta_is_enc = KeyHolder.step_4b(
            beta, self.l_maximum_bit_length, self.scheme_dgk
        )

        d_enc.randomize()
        for b in beta_is_enc:
            b.randomize()
        await self.communicator.send(
            self.other_party,
            (d_enc, beta_is_enc),
            msg_id=f"step_4b_session_{session_id}",
        )
        c_is_enc = await self.communicator.recv(
            self.other_party, msg_id=f"step_4i_session_{session_id}"
        )

        # step 4j
        delta_b = KeyHolder.step_4j(c_is_enc, self.scheme_dgk)

        # step 5
        zeta_1_enc, zeta_2_enc, delta_b_enc = KeyHolder.step_5(
            z_plain, self.l_maximum_bit_length, delta_b, self.scheme_paillier
        )

        zeta_1_enc.randomize()
        zeta_2_enc.randomize()
        delta_b_enc.randomize()
        await self.communicator.send(
            self.other_party,
            (zeta_1_enc, zeta_2_enc, delta_b_enc),
            msg_id=f"step_5_session_{session_id}",
        )

    async def make_and_send_encryption_schemes(
        self,
        session_id: int = 1,
        key_length_paillier: int = 2048,
        v_bits_dgk: int = 160,
        n_bits_dgk: int = 2048,
    ) -> None:
        """
        Initialize Paillier and DGK encryption schemes if they don't
        already exist and sends public keys to Alice.

        :param session_id: integer to distinguish between session
        :param key_length_paillier: key length paillier
        :param v_bits_dgk: number of bits DGK private keys $v_p$ and $v_q$
        :param n_bits_dgk: number of bits DGK public key $n$
        :raise ValueError: raised when communicator is not propertly configured.
        """
        if self.communicator is None:
            raise ValueError("Communicator not properly initialized.")

        if self._scheme_paillier is None:
            self._scheme_paillier = Paillier.from_security_parameter(
                key_length=key_length_paillier
            )

        if self._scheme_dgk is None:
            self._scheme_dgk = DGK.from_security_parameter(
                v_bits=v_bits_dgk,
                n_bits=n_bits_dgk,
                u=next_prime(1 << (self.l_maximum_bit_length + 2)),
                full_decryption=False,
            )

        await self.communicator.send(
            self.other_party,
            (self.scheme_paillier, self.scheme_dgk),
            msg_id=f"schemes_session_{session_id}",
        )

    def _start_randomness_generation(self) -> None:
        """
        Boot randomness generation in both encryption schemes.
        """
        self.scheme_paillier.boot_randomness_generation(3)
        self.scheme_dgk.boot_randomness_generation(self.l_maximum_bit_length + 1)

    @staticmethod
    def step_2(
        z_enc: PaillierCiphertext, l: int, scheme_paillier: Paillier
    ) -> tuple[int, int]:
        r"""
        $B$ decrypts $[[z]]$, and computes $\beta = z \mod 2^l$.

        :param z_enc: Encrypted value of $z$: $[[z]]$.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x, y$ that will be
            given as input to this method.
        :param scheme_paillier: Paillier encryption scheme.
        :return: Tuple containing as first entry the plaintext value of $z$.
            The second entry is the value $\beta = z \mod 2^l$.
        """
        z = cast(int, scheme_paillier.decrypt(z_enc, apply_encoding=False))
        return z, z % (1 << l)

    @staticmethod
    def step_4a(
        z: int, scheme_dgk: DGK, scheme_paillier: Paillier, l: int
    ) -> DGKCiphertext:
        r"""
        $B$ computes the encrypted bit $[d]$ where $d = (z < (N - 1)/2)$ is the bit informing $A$
        whether a carryover has occurred.

        :param z: Plaintext value of $z$.
        :param scheme_dgk: DGK encryption scheme.
        :param scheme_paillier: Paillier encryption scheme.
        :param l: maximum bit length used to constrain variables.
        :return: Encrypted value of the bit $d = (z < (N - 1)/2)$: $[d]$.
        """
        assert scheme_dgk.public_key.u > (1 << (l + 2))
        return scheme_dgk.unsafe_encrypt(
            int(z < (scheme_paillier.public_key.n - 1) // 2),
            apply_encoding=False,
        )

    @staticmethod
    def step_4b(beta: int, l: int, scheme_dgk: DGK) -> list[DGKCiphertext]:
        r"""
        $B$ computes the encrypted bits $[\beta_i], 0 \leq i < l$ to $A$.

        :param beta: The value $\beta$ from step 2.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x, y$ that will be
            given as input to this method.
        :param scheme_dgk: DGK encryption scheme.
        :return: List containing the encrypted values of the bits $\beta_i$: $[\beta_i],
            0 \leq i < l$ to $A$.
        """
        return [
            scheme_dgk.unsafe_encrypt(bit, apply_encoding=False)
            for bit in to_bits(beta, l)
        ]

    @staticmethod
    def step_4j(c_is_enc: list[DGKCiphertext], scheme_dgk: DGK) -> int:
        r"""
        $B$ checks whether one of the numbers $c_i$ is decrypted to zero. If he finds one,
        $\delta_B \leftarrow 1$, else $\delta_B \leftarrow 0$.

        :param c_is_enc: List containing the encrypted values of the bits $c_i$: $[c_i],
            0 \leq i < l$.
        :param scheme_dgk: DGK encryption scheme.
        :return: Value $\delta_B$.
        """
        return int(
            any(
                map(
                    scheme_dgk.is_zero,
                    c_is_enc,
                )
            )
        )

    @staticmethod
    def step_5(
        z: int, l: int, delta_b: int, scheme_paillier: Paillier
    ) -> tuple[PaillierCiphertext, PaillierCiphertext, PaillierCiphertext]:
        r"""
        $B$ computes $\zeta_1 = z \div 2^l$ and encrypts it to $[[\zeta_1]]$ and computes
        $\zeta_2 = (z + N) \div 2^l$ and encrypts it to $[[\zeta_2]]$. $B$ also encrypts
        $\delta_B$ to $[[\delta_B]]$.

        :param z: Plaintext value of $z$.
        :param l: Fixed value, such that $0 \leq x,y < 2^l$, for any $x, y$ that will be
            given as input to this method.
        :param delta_b: The value $\delta_B$ from step 4j.
        :param scheme_paillier: Paillier encryption scheme.
        :return: A tuple with the first entry being the encrypted value of $\zeta_1$: $[[\zeta_1]]$.
            The second entry is the encrypted value of $\zeta_2$: $[[\zeta_2]]$.
            The third entry is the encrypted value of $\delta_B$: $[[\delta_B]]$.
        """
        # We use the fix mentioned in the paper here, to also make this work in case of overflow
        zeta_1_enc = scheme_paillier.unsafe_encrypt(z // (1 << l), apply_encoding=False)
        zeta_2_enc = (
            scheme_paillier.unsafe_encrypt(
                (z + scheme_paillier.public_key.n) // (1 << l),
                apply_encoding=False,
            )
            if z < (scheme_paillier.public_key.n - 1) // 2
            else scheme_paillier.unsafe_encrypt(z // (1 << l), apply_encoding=False)
        )
        return (
            zeta_1_enc,
            zeta_2_enc,
            scheme_paillier.unsafe_encrypt(delta_b, apply_encoding=False),
        )



"""
Pytest fixtures.
"""

# pylint: disable=redefined-outer-name
from __future__ import annotations

import asyncio
import warnings
from collections.abc import Iterator
from typing import Any

import pytest

from tno.mpc.communication import Pool, Serialization
from tno.mpc.encryption_schemes.dgk import DGK
from tno.mpc.encryption_schemes.paillier import Paillier
from tno.mpc.encryption_schemes.utils import next_prime

from tno.mpc.protocols.secure_comparison import Communicator, Initiator, KeyHolder

COMPARISON_BIT_LENGTH = 16
PAILLIER_MIN_KEY_LENGTH = 1024


@pytest.fixture
def _error_on_encryption_scheme_warnings() -> Iterator[None]:
    """
    Wrapper that turns encryption scheme warnings into errors.

    :return: Context with custom warningfilters.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("error", ".*ciphertext", UserWarning)
        warnings.filterwarnings("error", ".*randomness", UserWarning)
        yield


@pytest.fixture
def paillier_strict(_error_on_encryption_scheme_warnings: None) -> Iterator[Paillier]:
    """
    Yields a Paillier encryption scheme and properly shuts it down afterwards.

    This fixture will promote encryption scheme warnings into errors (e.g. it is "strict").

    :return: Paillier encryption scheme.
    """
    # _error_on_encryption_scheme_warnings fixture promotes selected warnings into errors
    paillier = Paillier.from_security_parameter(key_length=PAILLIER_MIN_KEY_LENGTH)
    yield paillier
    paillier.shut_down()


@pytest.fixture
def dgk_not_full_strict(_error_on_encryption_scheme_warnings: None) -> Iterator[DGK]:
    """
    Yields a DGK encryption scheme without full decryption and properly shuts
    it down afterwards.

    This fixture will promote encryption scheme warnings into errors (e.g. it is "strict").

    :return: DGK encryption scheme without full decryption.
    """
    # _error_on_encryption_scheme_warnings fixture promotes selected warnings into errors
    u = next_prime(1 << (COMPARISON_BIT_LENGTH + 2))
    dgk = DGK.from_security_parameter(v_bits=20, n_bits=128, u=u, full_decryption=False)
    yield dgk
    dgk.shut_down()


@pytest.fixture
def alice() -> Initiator:
    """
    Initiator for the secure comparison protocol with full DGK decryption.

    :return: Initiator.
    """
    return make_initiator(COMPARISON_BIT_LENGTH)


@pytest.fixture
def bob_not_full(paillier_strict: Paillier, dgk_not_full_strict: DGK) -> KeyHolder:
    """
    Keyholder for the secure comparison protocol without full DGK decryption.

    :param paillier_strict: Paillier scheme.
    :param dgk_not_full_strict: DGK scheme without full decryption.
    :return: Keyholder.
    """
    return make_keyholder(
        COMPARISON_BIT_LENGTH,
        paillier_strict,
        dgk_not_full_strict,
    )


@pytest.fixture
def playerpair_with_pool_communication(
    http_pool_duo: tuple[Pool, Pool],
    paillier_strict: Paillier,
    dgk_not_full_strict: DGK,
) -> tuple[Initiator, KeyHolder]:
    """
    Fixture for a secure comparison player pair. The players communicate
    through an http pool.

    :param http_pool_duo: Communication pools.
    :param paillier_strict: Paillier scheme.
    :param dgk_not_full_strict: DGK scheme.
    :return: Initiator and KeyHolder for secure comparison. The players
        communicate through http.
    """
    communicator_initiator = http_pool_duo[0]
    communicator_keyholder = http_pool_duo[1]
    initiator_id = next(iter(communicator_keyholder.pool_handlers))
    keyholder_id = next(iter(communicator_initiator.pool_handlers))

    initiator = make_initiator(
        COMPARISON_BIT_LENGTH,
        communicator=communicator_initiator,
        other_party=keyholder_id,
    )
    keyholder = make_keyholder(
        COMPARISON_BIT_LENGTH,
        paillier_strict,
        dgk_not_full_strict,
        communicator=communicator_keyholder,
        other_party=initiator_id,
    )
    return initiator, keyholder


@pytest.fixture
def playerpair_with_dict_communication(
    paillier_strict: Paillier,
    dgk_not_full_strict: DGK,
) -> tuple[Initiator, KeyHolder]:
    """
    Fixture for a secure comparison player pair. The players communicate
    through a dictionary.

    :param paillier_strict: Paillier scheme.
    :param dgk_not_full_strict: DGK scheme.
    :return: Initiator and KeyHolder for secure comparison. The players
        communicate through a dictionary.
    """
    communicator = DictionaryCommunicator()

    initiator = make_initiator(
        COMPARISON_BIT_LENGTH,
        communicator=communicator,
    )
    keyholder = make_keyholder(
        COMPARISON_BIT_LENGTH,
        paillier_strict,
        dgk_not_full_strict,
        communicator=communicator,
    )
    return initiator, keyholder


class DictionaryCommunicator:
    """
    Simple communicator based on a dictionary.
    """

    def __init__(self) -> None:
        """
        Initializes a dictionary communicator.
        """
        self.dict: dict[str, Any] = {}

    async def send(self, party_id: str, message: Any, msg_id: str) -> None:
        """
        Send a message to the other party.

        :param party_id: ID of the receiving party.
        :param message: Message to be send.
        :param msg_id: ID of the message.
        """
        del party_id
        message = Serialization.pack(message, msg_id=msg_id, use_pickle=False)
        self.dict[msg_id] = message

    async def recv(self, party_id: str, msg_id: str) -> Any:
        """
        Receive a message from the other party.

        :param party_id: ID of the receiving party.
        :param msg_id: ID of the message.
        :return: The contents of the message.
        """
        del party_id
        for _ in range(1_000):
            if not msg_id in self.dict:
                await asyncio.sleep(0.001)
        obj = self.dict.pop(msg_id)
        return Serialization.unpack(obj)[1]


def make_initiator(
    bit_length: int,
    communicator: Communicator | None = None,
    other_party: str = "",
) -> Initiator:
    """
    Helper function to create an initiator.

    :param bit_length: Maximum bit length of plaintext inputs to the secure
        comparison.
    :param communicator: Communication instance.
    :param other_party: ID of the other party.
    :return: Initiator.
    """
    return Initiator(
        l_maximum_bit_length=bit_length,
        communicator=communicator,
        other_party=other_party,
    )


def make_keyholder(
    bit_length: int,
    paillier_scheme: Paillier,
    dgk_scheme: DGK,
    communicator: Communicator | None = None,
    other_party: str = "",
) -> KeyHolder:
    """
    Helper function to create a keyholder.

    :param bit_length: Maximum bit length of plaintext inputs to the secure
        comparison.
    :param paillier_scheme: Paillier scheme.
    :param dgk_scheme: DGK scheme.
    :param communicator: Communication instance.
    :param other_party: ID of the other party.
    :return: Keyholder.
    """
    return KeyHolder(
        l_maximum_bit_length=bit_length,
        scheme_paillier=paillier_scheme,
        scheme_dgk=dgk_scheme,
        communicator=communicator,
        other_party=other_party,
    )



""" Functions that don't belong to one of the players"""

from __future__ import annotations


def to_bits(integer: int, bit_length: int) -> list[int]:
    """
    Convert a given non-negative integer to a list of bits, with the least significant bit
    first, and the most significant bit last.

    :param integer: Integer to be converted to bits.
    :param bit_length: Amount of bits to which the integer should be converted.
    :return: Bit representation of the integer in bit_length bits. Least significant bit first,
        most significant last.
    """
    assert integer < (1 << bit_length)
    bits = [0 for _ in range(bit_length)]
    for bit_index in range(bit_length):
        bits[bit_index] = integer & 1
        integer >>= 1
    return bits


def from_bits(bits: list[int]) -> int:
    """
    Convert a set of bits, least significant bit first to a
    non-negative integer.

    :param bits: List of bits, least significant bit first.
    :return: Integer representation of the bits.
    """
    base = 1
    integer = 0
    for bit in bits:
        if bit == 1:
            integer += base
        base *= 2
    return integer



"""
Integration tests for the tno.mpc.protocols.secure_comparison library.
"""

from __future__ import annotations

import asyncio

import pytest

from tno.mpc.encryption_schemes.paillier import Paillier, PaillierCiphertext

from tno.mpc.protocols.secure_comparison import Initiator, KeyHolder


async def run_interactive_comparison_with_pool(
    x_enc: PaillierCiphertext,
    y_enc: PaillierCiphertext,
    initiator: Initiator,
    keyholder: KeyHolder,
) -> PaillierCiphertext:
    """
    Perform integral comparison protocol including communication
    between Initiator and KeyHolder for integration test.

    :param x_enc: Encryption of first input.
    :param y_enc: Encryption of second input.
    :param initiator: Initiator of the secure comparison.
    :param keyholder: Keyholder in the secure comparison.
    :return: Encryption of x <= y.
    """
    task_initiator = asyncio.create_task(
        initiator.perform_secure_comparison(x_enc, y_enc)
    )
    task_keyholder = asyncio.create_task(keyholder.perform_secure_comparison())

    x_leq_y_enc, _ = await asyncio.gather(*(task_initiator, task_keyholder))
    return x_leq_y_enc


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "x, y, expected_result", ((0, 1, True), (0, 0, True), (1, 0, False))
)
async def test_pool(
    x: int,
    y: int,
    expected_result: bool,
    paillier_strict: Paillier,
    playerpair_with_pool_communication: tuple[Initiator, KeyHolder],
) -> None:
    r"""
    Validates integration with tno.mpc.communication.Pool as communicator.

    Computes $x \leq y$ using a Pool object for communication.

    :param x: First input to secure comparison.
    :param y: Second input to secure comparison.
    :param expected_result: bool($x\leq y$)
    :param paillier_strict: Paillier scheme.
    :param playerpair_with_pool_communication: Initiator-keyholder pair that
        is configured with communication.
    """
    x_enc = paillier_strict.unsafe_encrypt(x)
    y_enc = paillier_strict.unsafe_encrypt(y)
    initiator, keyholder = playerpair_with_pool_communication

    x_leq_y_enc = await run_interactive_comparison_with_pool(
        x_enc, y_enc, initiator, keyholder
    )
    x_leq_y = paillier_strict.decrypt(x_leq_y_enc)

    assert bool(x_leq_y) == expected_result
